#include <stdio.h>
#include "segfont.h"

typedef struct {	/* Buffer that holds rows to be output */
   LINE *Buffer;	/* Array of pointers to output columns */
   FILE *InFile;	/* File to read chars from */
   /* How many output char rows to skip before starting column printing
      (to allow for bottom-adjusting when printing down) */
   int StartPoints[MAX_COLS];
   int CharsAcross,	/* How many chars are to be printed across display */
   Backwards;	/* Whether strings are to be reversed and bottom adjusted */
} ConvBuf;

char *CLine;

#ifdef __MSDOS__
#include <alloc.h>
#endif

#ifdef _M_SYSV
#define M_SYS5
#endif

#ifdef M_SYS5
#include <string.h>
#include "segpr.h"
int FillDown(ConvBuf *RowBuf,int CharWidths[],int Spacing);
void SetStarts(ConvBuf *RowBuf,int CharWidths[],int Spacing);
void PrintCols(ConvBuf *RowBuf,int Size,int Width,CharDesc Font[],int Dim[][2],
 int Spacing,char *OutBuf[],int CharWidths[],int UseTabs);
#else
#include <strings.h>
#endif

/* PrintDown: Reads lines of input from InFile and prints them vertically in
   segment font.
FILE *InFile;	  Input file 
int Backwards,
CharWidths[],	  Widths of chars when expanded 
Size,		  Number of chars per segment to use 
Width,		  Char width of output device 
Spacing,	  Number of spaces between output chars 
Dim[][2];	  First and last cols used by each char 
CharDesc Font[];	  Font to use in printing 
int DoFF;	  Whether to print a formfeed after each set of lines 
int UseTabs;	  Whether to use tabs instead of spaces wherever possible 
char *S;	  String to print instead of a file 
*/
void PrintDown(FILE *InFile,int Backwards,int Size,int Width,CharDesc Font[],
int Dim[][2],int Spacing,int CharWidths[],int DoFF, int UseTabs, char *S)
{
   int CharsAcross;	/* Number of chars that can be printed across page */
   ConvBuf ColBuf;	/* Input row buffer */
   int FileDone;	/* Whether end of file has been reached */
   /* Array of output lines that seg chars are written to */
   char *OutBuf[MAX_COLS];
   int NumRows;		/* Number of rows in output buffer */
   int I;

   /* Two * Size buffer rows for the variable-size seg char rows, 3 for the
      fixed size rows, and Spacing buffer rows for spaces between seg chars */
   NumRows = Size * 2 + 3 + Spacing;
   /* Number of output chars across is as many as will fit (plus one display
      space between output chars) in the width given minus one space (to avoid
      problems with screen/printer wraparound) plus one space (since no space
      is needed after last char) */
   CharsAcross = Width / (Size * 2 + 4);
   /* Set up memory for output buffer lines */
   for (I = 0; I < NumRows; I++)
      OutBuf[I] = (char *) CallocT(Width,sizeof(char));
   /* Allocate memory for input buffer columns */
   ColBuf.Buffer = (LINE *) CallocT(CharsAcross * MAX_LINE,sizeof(char));
   ColBuf.InFile = InFile;
   ColBuf.CharsAcross = CharsAcross;
   ColBuf.Backwards = Backwards;
   CLine = S;
   /* Get first buffer of strings */
   FileDone = FillDown(&ColBuf,CharWidths,Spacing);
   while (!FileDone) {
      /* Print next row in seg font */
      PrintCols(&ColBuf,Size,Width,Font,Dim,Spacing,OutBuf,CharWidths,UseTabs);
      if (DoFF)
         putchar('\f');		/* Print a formfeed */
      /* Get next row from buffer */
      FileDone = FillDown(&ColBuf,CharWidths,Spacing);
   }
}

/*
ConvBuf *ColBuf;	  Input column buffer
int Size,	  Number of chars per segment to use
Width,		  Char width of output device
Spacing,	  Number of spaces between output chars
Dim[][2];	  First and last cols used by each char
CharDesc Font[];	  Font to use in printing
char *OutBuf[];		  Output line buffer
int CharWidths[];	  Widths of output chars
int UseTabs;	  whether to use tabs instead of spaces wherever possible
*/
void PrintCols(ConvBuf *ColBuf,int Size,int Width,CharDesc Font[],int Dim[][2],
int Spacing, char *OutBuf[],int CharWidths[],int UseTabs)
{
   /* Count for each column telling how many more rows to print before that
      the last character written to the buffer for that column is completed
      and the next must be written */
   int RowCount[MAX_COLS];
   /* Pointers to the next char to be printed in each column of the input
      buffer */
   char *ColPoints[MAX_COLS];
   int NumUnfinished;	/* Number of columns which still have chars to print */
   int NumCols;		/* Number of columns */
   int NumRows;		/* Number of rows in output buffer */
   int RowNum,		/* Number of output buffer row being printed */
   ColNum;		/* Output char column that char is being written to */
   char Line[MAX_LINE];	/* Copy of output line to be compressed and printed */
   char InChar;		/* Character to write to buffer */
   int ColWidth;	/* Width of each output column */
   /* Font for a blank char */
   static CharDesc EmptyFont = {"     ","     ","     ","     ","     "};
   /* Char dimension spec indicating that all cols or rows should be printed */
   static int FullDim[] = {0,4};

   /* Column width is max width of an output char + 1 space between output char
      columns */
   ColWidth = (Size << 1) + 4;
   /* Number of rows is max height of an output char + num of spaces between
      chars */
   NumRows = (Size << 1) + 3 + Spacing;
   NumCols = ColBuf -> CharsAcross;
   for (ColNum = 0; ColNum < NumCols; ColNum++) {
      /* Copy starting points (number of display lines to skip before printing
         each char) to RowCount */
      RowCount[ColNum] = ColBuf -> StartPoints[ColNum];
      /* Set column pointers to the start of each input column */
      ColPoints[ColNum] = ColBuf -> Buffer[ColNum];
   }
   /* Clear and terminate output buffer lines */
   for (RowNum = 0; RowNum < NumRows; RowNum++) {
      memset(OutBuf[RowNum],' ',Width - 1);
      OutBuf[RowNum][Width-1] = '\0';
   }
   RowNum = 0;	/* Start printing to output buffer row 0 */
   NumUnfinished = NumCols;	/* Start with no columns finished */
   while (NumUnfinished) {
      for (ColNum = 0; ColNum < NumCols; ColNum++)
         if (!RowCount[ColNum]--) {
            InChar = *ColPoints[ColNum];
            if (InChar == '\0') {
               WriteChar(OutBuf,ColNum * ColWidth,0,Size,Spacing,0,EmptyFont,
                FullDim,FullDim,NumRows);
               RowCount[ColNum] = 32000;
               NumUnfinished--;
            }
            else {
               ColPoints[ColNum]++;
      /* Write char to buffer.  Turbo C likes the & in front of Font, cc
         does not; both compile correctly but issue warnings if other form
         is used. */
#ifdef __MSDOS__
               WriteChar(OutBuf,ColNum * ColWidth,RowNum,Size,Spacing,0,
                &Font[InChar],Dim[InChar],FullDim,NumRows);
#else
               WriteChar(OutBuf,ColNum * ColWidth,RowNum,Size,Spacing,0,
                Font[InChar],Dim[InChar],FullDim,NumRows);
#endif
               RowCount[ColNum] = CharWidths[InChar] + Spacing - 1;
            }
         }
      if (NumUnfinished) {
         strcpy(Line,OutBuf[RowNum]);
         CompressLine(Line,UseTabs);		/* Compress & print line */
         puts(Line);
         if (++RowNum == NumRows)
            RowNum = 0;
      }
   }
}

/* Fills buffer in ColBuf, reading strings and storing them in sequential
   columns of the buffer. */
int FillDown(ConvBuf *ColBuf,int CharWidths[],int Spacing)
{
   int Col;
   char *InLine;		/* Line read in */
   int LineLen;			/* Length of line read in */
   int GoodLine = FALSE;	/* Whether a non-null line was found */

   /* For each column in the buffer */
   for (Col = 0; Col < ColBuf -> CharsAcross; Col++) {
      InLine = ColBuf -> Buffer[Col];
      /* If no more input, make line null */
      if (ColBuf->InFile) {
	  if (feof(ColBuf -> InFile) ||
	   fgets(InLine,MAX_LINE,ColBuf -> InFile) == NULL)
	     *InLine = '\0';
      }
      else {
	 strcpy(InLine,CLine);
	 CLine[0] = '\0';
      }
      LineLen = strlen(InLine);
      GoodLine |= LineLen > 0;
      if (LineLen && InLine[LineLen - 1] == '\n')
         InLine[LineLen - 1] = '\0';
      if (ColBuf -> Backwards)		/* Reverse char order in line */
         strrev(InLine);
   }
   if (GoodLine)
      if (ColBuf -> Backwards)
         SetStarts(ColBuf,CharWidths,Spacing);
      else
         SetInts(ColBuf -> StartPoints,ColBuf -> CharsAcross,0);
   return(!GoodLine);
}

void SetStarts(ConvBuf *ColBuf,int CharWidths[],int Spacing)
{
   int ColNum;
   char *Col;
   int Rows,InChar;
   int Longest = 0;

   for (ColNum = 0; ColNum < ColBuf -> CharsAcross; ColNum++) {
      Rows = 0;
      Col = ColBuf -> Buffer[ColNum];
      for (InChar = 0; Col[InChar] != '\0'; InChar++)
         Rows += Spacing + CharWidths[Col[InChar]];
      ColBuf -> StartPoints[ColNum] = Rows;
   }
   for (ColNum = 0; ColNum < ColBuf -> CharsAcross; ColNum++)
      Longest = Max(Longest,ColBuf -> StartPoints[ColNum]);
   for (ColNum = 0; ColNum < ColBuf -> CharsAcross; ColNum++)
      ColBuf -> StartPoints[ColNum] = Longest - ColBuf -> StartPoints[ColNum];
}
