#include <ctype.h>
#include <stdio.h>
#include "segfont.h"

#ifdef _M_SYSV
#define M_SYS5
#endif

#ifdef M_SYS5
#include <string.h>
#include <stdlib.h>
int ConvInt(char **S);
int GetValue(char **String,int *Value,char FlagChar,int MinVal);
void PrintStrings(char *Strings[]);
#else
#include <strings.h>
#endif

char *HelpScreen[] = {
"segfont: Write strings as 25-segment chars using character graphics.",
"Usage:  segfont [-hkpdb] [-c<char>] [-f<file>] [-g<gap>]",
"        [-r<rotates>] [-s<size>] [-w<width>] [string ...]",
"   If no strings are given on command line, standard input is read.",
"   -b (backwards): Reverse input strings before printing.",
"   -c : Replace the default segment chars with <char>.  If <char> is ",
"not given, output chars will be built from the char they are representing.",
"   -d : Print strings down (vertically) in adjacent columns.  If -b is also",
"given, output will be bottom-adjusted (strings will start at the bottom).",
"   -e : Do not use tabs in output.",
"   -f : Use 25-segment font given in <file> instead of internal font.",
"See the file \"segfont.def\" for an example.",
"   -g : Set gap (number of spaces between characters) to <gap>.",
"The default is to set the gap to the character size.",
"   -h : Print this help screen.",
"   -k : Kill proportional spacing; use fixed spacing instead.",
"   -p : Issue a page eject (formfeed, ^L) after each set of columns when",
"printing down.  It has no effect if -d is not also given.",
"   -r : Rotate output chars clockwise 90*<rotates> degrees.  ",
"<rotates> should be between one and three.",
"   -s : Set char size (the number of segment chars that each variable-size",
"segment is constructed from).  It must be at least 1.  The default is 3.",
"   -w : Set page width to <width> chars.  The default is 80.",
NULL};

/* Sets program parameters according to the values given in Params. */
int InterpretFlags(int Num,char *Params[],int *Rotates,int *Backwards,
int *Size,int *Width,int *Down,char *SegChar, char **FontFile,int *DoFF,
int *Spacing,int *FixedSpace,int *UseTabs)
{
   int I;
   /* Leading char from a string in Params, specifying which flag it is. */
   char FlagChar;
   char *CurPar;	/* Current string from Params */
   int Error;	/* Whether an error was encountered in the current param */
   int BadFlag = FALSE;	/* Whether any errors occured */

   for (I = 0; I < Num; I++) {
      CurPar = Params[I];
      if (*CurPar == '\0') {
         fputs("Null flag given.\n",stderr);
         BadFlag = TRUE;
         continue;
      }
      while (*CurPar != '\0') {
         FlagChar = LowCase(*CurPar);
         CurPar++;
         Error = FALSE;
         switch (FlagChar) {
            case 'b':	*Backwards = TRUE;
         		break;
            case 'c':	if (*CurPar == '\0')
            		   *SegChar = ' ';
            		else {
            		   *SegChar = *CurPar;
                   	   /* Increment CurPar to account for
                              char given with -c flag */
            		   CurPar++;
           		}
			/* Let graphic chars be used but not control chars */
         		Error = !isprint(*SegChar) && (unsigned) *SegChar < 128;
         		if (Error)
         		   fputs("Bad char given with -c argument.\n",stderr);
         		break;
            case 'd':	*Down = TRUE;
         		break;
            case 'e':	*UseTabs = FALSE;
            		break;
            case 'f':	*FontFile = CurPar;
         		Error = *CurPar == '\0';
         		/* Make CurPar point to a null char to indicate that
         		   there are no more flags in Curpar */
         		CurPar = "\0";
         		if (Error)
         		   fputs("No filename given with -f arguement.\n",
         		    stderr);
         		break;
            case 'g':	*Spacing = -1;
            		Error = GetValue(&CurPar,Spacing,FlagChar,0);
         		break;
            case 'h':	PrintStrings(HelpScreen);
         		Error = TRUE;
         		break;
	    case 'k':	*FixedSpace = TRUE;
			break;
            case 'p':	*DoFF = TRUE;
         		break;
            case 'r':	*Rotates = 0;
            		Error = GetValue(&CurPar,Rotates,FlagChar,1);
         		break;
            case 's':	*Size = 0;
            		Error = GetValue(&CurPar,Size,FlagChar,1);
         		break;
            case 'w':	*Width = 0;
            		Error = GetValue(&CurPar,Width,FlagChar,1);
         		break;
            default:	fprintf(stderr,"Bad flag: %c.\n",FlagChar,1);
         		Error = TRUE;
         }
         BadFlag |= Error;
      }
   }
   return(BadFlag);
}

int ConvInt(char **S)
/* Converts the digit chars pointed to by *S to an integer and returns it.
   No error occurs on overflow.  *S is moved to the next character beyond
   the last digit char.  If **S is not a digit char, 0 is returned and *S
   is not changed. */
{
   int Temp;

   for (Temp = 0; **S != '\0' && '0' <= **S && **S <= '9'; (*S)++)
      Temp = Temp * 10 + (unsigned char) **S - '0';
   return(Temp);
}

/* Converts *String to an integer value and assigns it to Value.  If no digits
   are given in *String, the value of *Value is not changed.  If the final
   value of *Value is less than MinVal, an error message is printed including
   character FlagChar, and TRUE is returned; otherwise *String is moved beyond
   the last digit char and FALSE is returned.  Since *Value is not changed if
   no digit chars are found, if the original value of *Value is less than
   MinVal, the error message will be printed if no digit chars are found.
*/
int GetValue(char **String,int *Value, char FlagChar, int MinVal)
{
   int Temp;		/* Temp storage of string conversion value */
   char *StartChar;	/* Pointer to original start of string */

   StartChar = *String;
   Temp = ConvInt(String);
   if (StartChar != *String)
      *Value = Temp;
   if (*Value < MinVal) {
      fprintf(stderr,"Bad value given with -%c flag.\n",FlagChar);
      return(TRUE);
   }
   else
      return(FALSE);
}

/* Separates argv into flags and filenames.  Flags and Files are malloced to
   be arrays of (argc) char pointers.  Pointers to flags (arguments that begin
   with one of the characters in FlagSpecs) are put in Flags,with the flag
   specifier removed from the front; pointers to all other argument are copied
   to Files.  The pointer in Flags and Files following the last pointer copied
   is set to NULL. The number of flags found is returned.  argv[0] is ignored.
*/
int SepFlags(int argc, char *argv[], char *FlagSpecs, char ***Flags,
char ***Files)
{
   int I;
   char **FlgTmp,**FilTmp;	/* Temporary flag and filename array ptrs */
   int NumFlgs = 0,NumFls = 0;	/* Number of flags and filenames found */

   FlgTmp = (char **) calloc(argc,sizeof(char *));
   FilTmp = (char **) calloc(argc,sizeof(char *));
   for (I = 1; I < argc; I++)	/* For each argument */
      if (strchr(FlagSpecs,*argv[I]) == NULL)
         FilTmp[NumFls++] = argv[I];
      else
         FlgTmp[NumFlgs++] = argv[I] + 1;
   FilTmp[NumFls] = NULL;
   FlgTmp[NumFlgs] = NULL;
   *Flags = FlgTmp;	/* Copy temp pointers to arrays to parameters */
   *Files = FilTmp;
   return(NumFlgs);
}

/* Prints an array of strings, one per line, until a terminating NULL pointer
   is found */
void PrintStrings(char *Strings[])
{
   int I;

   for (I = 0; Strings[I] != NULL; I++)
      puts(Strings[I]);
}
