/* 
*   SEGFONT: Write strings as 25-segment chars using character graphics.
*   1987     John H. DuBois III (john@armory.com)
*   88/10/22 rev
*   91/05/04 ported to XENIX
*   93/04/13 Modified to support strings given on command line,
*            to be more like banner programs
*/
#include "segfont.h"

#ifdef __MSDOS__
#include <alloc.h>
#include <process.h>
#define FLAG_SPEC "-/"	/* Allow both '-' and '/' to specify flags */
#else
#define FLAG_SPEC "-"
#endif

#ifdef _M_SYSV
#define M_SYS5
#endif

#ifdef M_SYS5
#include <string.h>
#endif

#include <stdio.h>

#ifdef M_SYS5
int SepFlags(int argc,char *argv[],char *FSpecs,char ***Flags,char ***Files);
int InterpretFlags(int Num,char *Params[],int *Rotates,int *Backwards,
 int *Size,int *Width,int *Down,char *SegChar,char **FontFile,int *DoFF,
 int *Spacing,int *FixedSpace,int *UseTabs);
void GenerateFont(long FontData[],char *Segments[],CharDesc Font[]);
void RotateFont(long FontData[],int Rotates);
int ReadFont(char *FontFile,long FontData[],char *Segments[]);
void PrintAcross(FILE *InFile,int Backwards,int Size,int Width,CharDesc Font[],
 int Dim[][2],int Spacing,int CharWidths[],int UseTabs,char *S);
void PrintDown(FILE *InFile,int Backwards,int Size,int Width,CharDesc Font[],
 int Dim[][2],int Spacing,int CharWidths[],int DoFF,int UseTabs,char *S);
void ConvertChars(CharDesc Font[],char SegChar);
void FindDim(long Font[],int Dim[][2],int Rows);
void FindWidths(int Dim[][2],int Size,int CharWidths[]);
#else
#include <strings.h>
#endif

long FontData[] = {	/* Default font data in packed form */
0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,
0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,
0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,
0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,
0x000000L,0x000000L,0x000000L,0x000000L,0x000000L,0x420080L,0x0000a0L,
0xa2a8a0L,0xa2aa8aL,0x26aac8L,0xa58188L,0x000040L,0x220040L,0x820100L,
0x0729c0L,0x022880L,0x040000L,0x002800L,0x400000L,0x040040L,0xac826aL,
0x008060L,0xa4082aL,0xa0882aL,0x022a80L,0xa1220aL,0xa8aa0aL,0x04004aL,
0xa8aa2aL,0x00892aL,0x401000L,0x041000L,0x010040L,0xa02800L,0x040100L,
0x42082aL,0x898a2aL,0x092300L,0xa9224aL,0xa8020aL,0x0c0300L,0xa8220aL,
0x08220aL,0xa88a0aL,0x08aa20L,0xa2008aL,0xa88020L,0x092240L,0xa80200L,
0x088360L,0x098320L,0xa8822aL,0x08224aL,0xa9822aL,0x092a2aL,0xa0aa0aL,
0x02008aL,0xa88220L,0x0c0240L,0x0d8220L,0x050140L,0x020140L,0xa4004aL,
0x220082L,0x010100L,0x820088L,0x000300L,0xa00000L,0x000100L,0x8b2000L,
0xa92200L,0x882000L,0xa48820L,0x8c2000L,0x022882L,0x822288L,0x0a2200L,
0x021000L,0x8a0080L,0x0300c0L,0x880200L,0x0aa800L,0x0c8800L,0x8a2000L,
0x082288L,0x222288L,0x082000L,0x210800L,0x022800L,0x8a0000L,0x0c0000L,
0xaa8000L,0x052800L,0x040140L,0x842000L,0x222082L,0x020080L,0x820888L,
0x000340L,0x000000L };

char *Segments[] = {	/* Default segment chars */
" -O- ",
"|\\|/|",
"O-O-O",
"|/|\\|",
" -O- " };

char *Comment = "@(#) segfont 1.1 96/07/03";

void main (int argc,char *argv[])
{
   /* How many 90 degree clockwise rotates to apply to chars */
   int Rotates = 0,
   /* Whether strings should be printed backwards (last char to first char) */
   Backwards = FALSE,
   Size = 3,
   Width = 80,		/* Width of output device, set to default */
   /* Whether strings should be printed down the page instead of across */
   Down = FALSE;
   /* Char to chance all segment chars to.  If not altered from null char,
      no change will be done. */
   char SegChar = '\0';
   char *FontFile = NULL;	/* Name of fontfile from -f arg (if given) */
   /* Command line args separated into flags and filenames */
   char **Flags,**Files;
   int NumFlags;	/* Number of flags given on command line */
   CharDesc Font[MAX_CHAR + 1];	/* Character font */
   int Dim[MAX_CHAR + 1][2];	/* Dimensions of characters */
   FILE *InFile;	/* Input file for translation to new font */
   /* Whether to do a formfeed after each group of cols written in down mode */
   int DoFF = FALSE;
   int Spacing = -1;	/* Number of spaces between output chars */
   /* Whether to use fixed spacing rather than proportional spacing */
   int FixedSpace = FALSE;
   /* Number of rows or cols (depending on whether Down printing is being done)
      used by each char in expanded form */
   int CharWidths[MAX_CHAR + 1];
   /* Whether tabs should be substituted for spaces in the output if it would
      reduce the number of display chars printed. */
   int UseTabs = TRUE;
   char S[MAX_LINE];
   int i;

   /* Separate command line args into flags and file names */
   NumFlags = SepFlags(argc,argv,FLAG_SPEC,&Flags,&Files);
   /* Use flags to set up translation parameters */
   if (InterpretFlags(NumFlags,Flags,&Rotates,&Backwards,&Size,&Width,
    &Down,&SegChar,&FontFile,&DoFF,&Spacing,&FixedSpace,&UseTabs))
      exit(2);
   ExTest(Width < Size * 2 + 2,
    "Output device not wide enough for one character.\n");
   ExTest(Width >= MAX_LINE,"Output too wide.\n");
   /* If font file name given on command line, read font file and set up
      font database */
   if (FontFile != NULL && ReadFont(FontFile,FontData,Segments))
      exit(2);	/* Exit if bad filename given */
   /* If file name given on command line, open it and use it for input.
      If more than one file name is given, all except the first are
      ignored. */
   InFile = stdin;
   if ((argc - NumFlags) > 1) {
      InFile = NULL;
      S[0] = '\0';
      for (i = NumFlags + 1; i < (argc - 1); i++) {
	 strcat(S,argv[i]);
	 strcat(S," ");
      }
      strcat(S,argv[i]);
   }
   if (Rotates)		/* Rotate packed font spec for each char */
      RotateFont(FontData,Rotates);
   /* Convert packed font data to segment char array representations
      for each char */
   GenerateFont(FontData,Segments,Font);
   if (FixedSpace)
      /* If fixed spacing, make each char occupy full character area */
      MakeFixed(Dim);
   else
      FindDim(FontData,Dim,Down);	/* Find character dimensions */
   FindWidths(Dim,Size,CharWidths);		/* Find char widths */
   if (SegChar != '\0')	/* If new segment char given, translate font to it */
      ConvertChars(Font,SegChar);
   if (Spacing == -1)	/* If spacing not specified, set spacing to Size */
      Spacing = Size;
   /* Traslate input chars to new font and send to stdout. */
   if (Down)
      PrintDown(InFile,Backwards,Size,Width,Font,Dim,Spacing,CharWidths,DoFF,
       UseTabs,S);
   else
      PrintAcross(InFile,Backwards,Size,Width,Font,Dim,Spacing,CharWidths,
       UseTabs,S);
}

/* If SegChar is a space, convert all non-space display chars in each Font
   entry to the same character as the input char which it is the entry for.
   Otherwise, convert all non-space display chars in Font to SegChar. */
void ConvertChars(CharDesc Font[],char SegChar)
{
   int I;
   char *Segs;
   int CharNum;

   Segs = (char *) Font;
   for (CharNum = 0; CharNum <= MAX_CHAR; CharNum++)
      for (I = 0; I < 25; I++) {
         if (*Segs != ' ')
            *Segs = SegChar == ' ' ? (char) CharNum : SegChar;
         Segs++;
      }
}

/* Set all elements of Dim so that each character occupies first to last row
   or col */
void MakeFixed(int Dim[][2])
{
   int I;

   for (I = 0; I <= MAX_CHAR; I++) {
      Dim[I][0] = 0;
      Dim[I][1] = 4;
   }
}

/* Find actual width of chars when expanded */
void FindWidths(int Dim[][2], int Size, int CharWidths[])
{
   int I,J;
   int RCWidth[5];	/* Width of each row or col */
   int Temp;

   RCWidth[0] = RCWidth[2] = RCWidth[4] = 1;
   RCWidth[1] = RCWidth[3] = Size;
   for (I = 0; I <= MAX_CHAR; I++) {
      Temp = 0;
      for (J = Dim[I][0]; J <= Dim[I][1]; J++)
         Temp += RCWidth[J];
      CharWidths[I] = Temp;
   }
}

/* Calls calloc() to allocate a block of memory for NElem elements of size
   Size and checks whether there was enough memory. Exits if not, otherwise
   returns a pointer to the block */
void *CallocT(unsigned NElem, unsigned Size)
{
void *Temp;

   Temp = (void *) calloc(NElem,Size);
   if (Temp == NULL) {
      fputs("Error: Not enough memory.\n",stderr);
      exit(2);
   }
   return(Temp);
}

#ifndef __MSDOS__
/* Reverses the order of the characters in S */
char *strrev(char *S)
{
   char C,*Last,*First;

   First = S;
   Last = S + strlen(S) - 1;
   while (Last > First) {
      C = *First;
      *(First++) = *Last;
      *(Last--) = C;
   }
   return(S);
}
#endif

#ifndef M_SYS5

/* Returns the length of the initial segment of S that consists entirely of
   chars that are not also in SearchChars; ie, returns an index to the first
   char in S that is also in SearchChars, and returns the length of S if no
   matching char found. */
unsigned strcspn(char *S,char *SearchChars)
{
   int I;

   for (I = 0; S[I] != '\0'; I++)
      if (strchr(SearchChars,S[I]) != NULL)
         return(I);
   return(I);
}

/* Sets Num bytes of memory, starting with Mem, to the value given by Byte */
void *memset(void *Mem,unsigned char Byte, unsigned Num)
{
   unsigned I;

   for (I = 0; I < Num; I++)
      ((char *)Mem)[I] = Byte;
   return(Mem);
}
#endif
