#include <stdio.h>
#include "segfont.h"

#ifdef __MSDOS__
#include <process.h>
#endif

#ifdef _M_SYSV
#define M_SYS5
#endif

#ifdef M_SYS5
#include <string.h>
#include <stdlib.h>
int GenFont(FILE *InFile,long FontData[],char *Template[]);
int CheckLine(char Template[],char Line[]);
void RowLimits(long FontData,int *Top,int *Bottom);
void ColLimits(long FontData,int *Left,int *Right);
long ReverseRows(long FontData);
long RowToCol(long FontData);
long ReverseLong(long Word,int NumBits);
void ReadLine(FILE *InFile,char *String,int Num);
long ConvertFont(CharDesc FontSpec);
#else
#include <strings.h>
int GenFont();
int CheckLine();
long ReverseRows();
long RowToCol();
long ReverseLong();
void RowLimits();
void ColLimits();
void ReadLine();
long ConvertFont();
#endif

/* Converts the bit-specified segments in FontData to Font */
void GenerateFont(long FontData[],char *Segments[],CharDesc Font[])
{
   int CharNum,Row,Col;
   long CharData;

   for (CharNum = 0; CharNum <= MAX_CHAR; CharNum++) {
      CharData = FontData[CharNum];
      for (Row = 0; Row < 5; Row++)
         for (Col = 4; Col >= 0; Col--) {
            Font[CharNum][Row][Col] = CharData & 1 ? Segments[Row][Col] : ' ';
            CharData >>= 1;
         }
   }
}

/* Rotatates the fonts in FontData in accordance with Rotates */
/* int Rotates;	Number of 90 deg. clkwise rotations to apply to font */
void RotateFont(long FontData[],int Rotates)
{
   int I;

   for (I = 0; I < MAX_CHAR; I++)
      switch (Rotates & 3) {
         case 1:	FontData[I] = RowToCol(ReverseRows(FontData[I]));
         		break;
         case 2:	FontData[I] = ReverseLong(FontData[I],25);
         		break;
         case 3:	FontData[I] = ReverseRows(RowToCol(FontData[I]));
         		break;
      }
}

/* Reverses the order of the rows in FontData and returns them */
long ReverseRows(long FontData)
{
   int I;
   long Temp = 0L;

   for (I = 0; I < 5; I++) {
      Temp = (Temp << 5)  |  (FontData & 0x1f);
      FontData >>= 5;
   }
   return(Temp);
}

/* Converts the rows in FontData to columns (row 0 becomes col 0, etc) and
   returns them */
long RowToCol(long FontData)
{
   long Temp = 0L,	/* Accumulator for result */
   Row,			/* Row bits being converted */
   Col;			/* Column bit of row being converted */
   int SourceRow,	/* Number of row being converted */
   SourceCol;		/* Number of col being converted */

   for (SourceRow = 0; SourceRow < 5; SourceRow++) {
      Row = FontData & 0x1f;	/* Get row to be converted */
      Col = 0L;			/* Clear col result */
      for (SourceCol = 0; SourceCol < 5; SourceCol++) {
      	/* Convert each bit of row to a col bit */
         Col = (Col << 5)  |  (Row & 1L);
         Row >>= 1;
      }
      Temp = (Temp << 1) | Col;	/* OR col bits into result */
      FontData >>= 5;		/* Move next row into low 5 bits */
   }
   return (Temp);
}

/* Reverse the low NumBits bits of Word and return them */
long ReverseLong(long Word, int NumBits)
{
   long Temp = 0L;
   int I;

   for (I = 0; I < NumBits; I++) {
      Temp = (Temp << 1) | (Word & 1L);
      Word >>= 1;
   }
   return(Temp);
}

/* Reads font segment file and converts to font data.
   Font segment file is a file containing fonts for some subset of chars
   0-127 in a 5x5 segment array form.  The file should begin with a template
   constisting of a 5x5 character array where each char in the array is the
   one that must be specified at that position if that segment is to be part
   of the char; this template is used for error checking.
   Each font specification should consist of the character number on a line,
   followed by a 5x5 character array where each character given indicates that
   that segment should be part of the char being specified.  The char given
   in each position must match the char in that position in the template.  If
   the segment is not to be part of the char, a space should be put in that
   position.  Trailing spaces need not be given.  Each font is written to
   the appropriate element of Font.
   Returns TRUE if errors occur; FALSE if not. */
int ReadFont(char *FontFile, long FontData[],char *Segments[])
{
   FILE *InFile;

   if ((InFile = fopen(FontFile,"r")) == NULL) {
      fputs("Error opening font file.",stderr);
      return(TRUE);
   }
   return(GenFont(InFile,FontData,Segments));	/* Generate font array */
}

/*
Reads InFile and converts data in it to Font.
char *Template[];	Chars for each segment
*/
int GenFont(FILE *InFile, long Font[], char *Template[])
{
   int I;
   CharDesc FontSpec;	/* Font to be converted */
   char InLine[80];	/* Line read in */
   int CharNum,		/* ASCII number of char that font is given for */
   LineNum,		/* Line number of font file being acted on */
   ColNum;		/* Column number of font spec that error occured on */

   for (I = 0; I < 5; I++)	/* Read template from first 5 lines of file */
      ReadLine(InFile,Template[I],5);
   LineNum = 6;
   while (!feof(InFile)) {
      do {
         fgets(InLine,80,InFile);	/* Skip blank lines */
         LineNum++;
      } while (!feof(InFile) && InLine[0] == '\0');
      if (!feof(InFile)) {
         CharNum = atoi(InLine);	/* Read character number given */
         if (!(0 < CharNum && CharNum <= MAX_CHAR)) {
            fprintf(stderr,"Bad character number at line %d: %d. (exiting)\n",
            LineNum - 1,CharNum);
            return(TRUE);
         }
         for (I = 0; I < 5; I++) {		/* Read font */
            ReadLine(InFile,FontSpec[I],5);
            LineNum++;
            if (feof(InFile)) {
               fputs("Unexpected end of file.\n",stderr);
               return(TRUE);
            }
            /* Compare font given to template */
            ColNum = CheckLine(Template[I],FontSpec[I]);
            if (ColNum)
               fprintf(stderr,"Bad char @ line %d; col %d, row %d of \
char font %d is \'%c\'; must be ' ' or \'%c\'.\n",
                LineNum - 1,ColNum,I + 1,CharNum,FontSpec[I][ColNum - 1],
                Template[I][ColNum - 1]);
         }
         Font[CharNum] = ConvertFont(FontSpec);	/* Convert font to long int */
      }
   }
   return(FALSE);
}

/* Converts an array of 5 x 5 chars to a long int.  If a character in the
   array is a space, the corresponding bit in the int is set to 0; if
   nonspace, the bit is set to 1.  The bits that correspond to each char
   location are as follows:
                   4  3  2  1  0
                   9  8  7  6  5
                  14 13 12 11 10
                  19 18 17 16 15
                  24 23 22 21 20
The high bits are set to 0. */
long ConvertFont(CharDesc FontSpec)
{
   long Temp;
   int Row,Col;

   Temp = 0L;
   for (Row = 4; Row >= 0; Row--)
      for (Col = 0; Col < 5; Col++) {
         Temp <<= 1;
         if (FontSpec[Row][Col] != ' ')
            Temp |= 1L;
      }
   return(Temp);
}

/* Compares the characters in Lines to the characters in Template.  If any
   nonspace char is given in Line, it must be the same as the char in the
   same position in Template.  Template and Line are arrays of 5 chars.
   If any bad chars are given, the char position (1-5) is returned;
   otherwise FALSE is returned. */
int CheckLine(char Template[], char Line[])
{
   int I;

   for (I = 0; I < 5; I++)
      if (Line[I] != ' ' && Line[I] != Template[I])
         return(I + 1);
   return(FALSE);
}

/* Reads a line from InFile of up to 80 chars; assigns the leftmost Num chars
   to String.  If the line read in is less than Num chars long, it is padded
   on the right with blanks. */
void ReadLine(FILE *InFile, char *String, int Num)
{
   char InString[80];
   int I;

   fgets(InString,80,InFile);
   strncpy(String,InString,Num);
   for (I = strcspn(String,"\n"); I < Num; I++)
      String[I] = ' ';
}

/* Find first & last row (if Rows is true) or first & last col of each font
   spec in Font.
   This information is used for proportional spacing when printing the
   characters.  Each number is between 0 and 4; 0 is the leftmost row or top
   column, 4 is the rightmost row or bottom column.
   An empty character (no segments used; either a space or an unspecified char)
   is taken to occupy the center row and center column.

Rows: Whether to find first & last columns used or first & last rows used.
*/
void FindDim(long Font[], int Dim[][2], int Rows)
{
   int I;
   long FontData;

   for (I = 0; I <= MAX_CHAR; I++) {
      FontData = Font[I];
      if (!FontData)
         /* If font is empty, set both dims to 2. */
         Dim[I][0] = Dim[I][1] = 2;
      else
         if (Rows)
            RowLimits(FontData,&Dim[I][0],&Dim[I][1]);
         else
            ColLimits(FontData,&Dim[I][0],&Dim[I][1]);
   }
}

/* Find first and last row occupied by FontData */
void RowLimits(long FontData, int *Top, int *Bottom)
{
   int Row,First = -1,Last;

   for (Row = 0; Row < 5; Row++) {
      if (FontData & 0x1fL) {	/* Check low 5 bits */
         Last = Row;		/* Update last occupied row */
         if (First == -1)	/* If first occupied row found, set First */
            First = Row;
      }
      FontData >>= 5;		/* Shift next row bits to low 5 bits */
   }
   *Top = First;
   *Bottom = Last;
}

/* Find first and last column occupied in FontData */
void ColLimits(long FontData,int *Left, int *Right)
{
   int Col,First = -1,Last;

   for (Col = 0; Col < 5; Col++) {
      if (FontData & 0x1084210L) {
       /* If low bit (leftmost col) of any row set */
         Last = Col;		/* Update last col */
         if (First == -1)	/* If first occupied col found, set First */
            First = Col;
      }
      FontData <<= 1; /* Shift next column bits into the low bit of each row */
   }
   *Left = First;
   *Right = Last;
}
