#include <stdio.h>
#include "segfont.h"

typedef struct {	/* Input buffer */
   char LineBuf[MAX_LINE];
   FILE *InFile;
   char  *CurChar;	/* Next char to be used */
   int Backwards;	/* Whether to reverse strings when read */
} InBuf;

#ifdef __MSDOS__
#include <alloc.h>
#include <process.h>
#endif

#ifdef _M_SYSV
#define M_SYS5
#endif

#ifdef M_SYS5
#include <string.h>
#include "segpr.h"
void PrintInRow(char InLine[],int NumChars,int Size,CharDesc Font[],
int Spacing,int Dim[][2],char *OutBuf[],int CharWidths[],int UseTabs);
int GetChars(InBuf *Buffer,int Width,int CharWidths[],char Line[],int Spacing);
void RestoreChar(InBuf *Buffer);
char NextChar(InBuf *Buffer);
void DeleteEndWhite(char *Line);
void SpcsToTabs(char *Line);
#else
#include <strings.h>
char NextChar();
#endif

/*
  Prints InLine in segment font using Font
char InLine[];		  Line of chars to be converted to output chars
CharDesc Font[];	  Font for chars
int NumChars,		  Number of input chars in InLine
  Number of display chars to print for segments with variable
   horizontal size
Size;
int Spacing,	  Number of spaces to put between chars
Dim[][2];	  First and last cols of font specs to use
char *OutBuf[];		  Output line buffer to write chars to
int CharWidths[];	  Widths of output chars
int UseTabs;	  Whether to use tabs instead of spaces wherever possible
*/
void PrintInRow(char InLine[], int NumChars, int Size, CharDesc Font[],
int Spacing, int Dim[][2], char *OutBuf[], int CharWidths[], int UseTabs)
{
   int CharNum;		/* Which char is being printed */
   int ColNum = 0;	/* Column to start printing next char at */
   char InChar;		/* Input char to print */
   /* Char dimensions struct used so that all cols of font will be printed */
   static int FullDim[] = {0,4};
   int NumRows,	/* Number of rows in buffer */
   Row;		/* Which row of buffer is being printed */

   /* Number of rows in buffer is max dimensions of a char plus one row
      for vertical spacing */
   NumRows = (Size << 1) + 4;
   for (CharNum = 0; CharNum < NumChars; CharNum++) {
      InChar = InLine[CharNum];
      /* Write char to buffer.  Turbo C likes the & in front of Font, cc
         does not; both compile correctly but issue warnings if other form
         is used. */
#ifdef __MSDOS__
      WriteChar(OutBuf,ColNum,0,Size,1,Spacing,&Font[InChar],
       FullDim,Dim[InChar],NumRows);
#else
      WriteChar(OutBuf,ColNum,0,Size,1,Spacing,Font[InChar],
       FullDim,Dim[InChar],NumRows);
#endif
      ColNum += CharWidths[InChar] + Spacing;
   }
   /* Terminate, compress, and print rows */
   for (Row = 0; Row < NumRows; Row++) {
      OutBuf[Row][ColNum] = '\0';
      CompressLine(OutBuf[Row],UseTabs);
      puts(OutBuf[Row]);
   }
}

/*
Writes an output character to OutBuf
int Size,	  Number of chars per segment to use
RowSpacing,ColSpacing,	  Number of spaces between output chars
RowDim[],ColDim[];	  First and last rows and cols used by char
CharDesc CharFont;	  Font to use in printing
char *OutBuf[];		  Output line buffer to write char to
int StartCol,		  Column in OutBuf to start writing to
RowNum,			  Row in OutBuf to start writing to
NumRows;		  Number of rows in OutBuf
*/
void WriteChar(char *OutBuf[], int StartCol, int RowNum, int Size,
int RowSpacing, int ColSpacing, CharDesc CharFont, int  RowDim[], int ColDim[],
int NumRows)
{
   int Row;	/* Which row of char font is being written */

   /* Write required segment rows */
   for (Row = RowDim[0]; Row <= RowDim[1]; Row++) {
      switch (Row) {
         case 0: case 2: case 4:
            /* Write segment rows with only horizontal components */
            WriteHoriz(OutBuf[RowNum],StartCol,ColDim,CharFont[Row],ColSpacing,
             Size);
            RowNum++;
            break;
         case 1: case 3:
            /* Write segment rows with vertical components (which will be
               written as multiple display rows if Size > 1) */
            WriteVert(OutBuf,StartCol,ColDim,CharFont[Row],ColSpacing,
             Row == 1 ? TRUE : FALSE,RowNum,NumRows,Size);
            RowNum += Size;
      }
      if (RowNum >= NumRows)
         RowNum -= NumRows;
   }
   /* Add spaces as required */
   for (Row = 0; Row < RowSpacing; Row++) {
      WriteHoriz(OutBuf[RowNum],StartCol,ColDim,"     ",ColSpacing,Size);
      if (++RowNum == NumRows)
         RowNum = 0;
   }
}

/*
Write rows of char font with vertical components (2nd and 4th rows) to OutBuf
char FontRow[];	 Font for row to be written
int ColLim[],	 First and last columns of font to use
StartCol,	 Column in OutBuf to start writing to
Upper,	 True if row to be written is the 2nd row; false if the 4th
Spacing,	 Number of spaces to write after the row
RowNum,		 Row of OutBuf to start writing to
NumRows,	 Number of rows in OutBuf
Size;		 Number of chars to write for multiple char segments
char *OutBuf[];	 Output buffer
*/
void WriteVert(char *OutBuf[], int StartCol, int ColLim[], char FontRow[],
int Spacing, int Upper, int RowNum, int NumRows, int Size)
{
   int CharCol;	/* Which column of font is being written */
   int Seg,	/* Counter for printing chars with both vert & diag compnts */
   /* Which row of this segment-row of the output char is being written */
   OutRow,
   LinePos;	/* Position in line buffer being written to */
   char *Line;	/* Line of OutBuf being written to */
   /* Position to print diagonal segments.  Tells the distance from the
      leftmost col that the left diagonal char should be printed and the
      distance from the rightmost col that the right diagonal char should
      be printed */
   int DiagCol;

   /* If upper segment row is being printed, first output row should have
      char in the right and leftmost cols; if lower, next to the middle col */
   DiagCol = Upper ? 0 : Size - 1;
   /* Draw segment row using the correct number of output rows */
   for (OutRow = 0 ; OutRow < Size; OutRow++) {
      Line = OutBuf[RowNum];
      if (++RowNum == NumRows)
         RowNum = 0;
      LinePos = StartCol;
      /* Draw the required columns */
      for (CharCol = ColLim[0]; CharCol <= ColLim[1]; CharCol++)
         switch (CharCol) {
            case 0: case 2: case 4:
		/* Store fixed position char */
		Line[LinePos++] = FontRow[CharCol];
		break;
            case 1:
               /* Print area occupied by second segment, placing segment in the
        	  correct position */
		for (Seg = 0; Seg < Size; Seg++)
		   Line[LinePos++] = Seg == DiagCol ? FontRow[1] : ' ';
		break;
	    case 3:
		/* Print area occupied by fourth segment, placing segment in
		   the correct position */
		for (Seg = Size - 1; Seg >= 0; Seg--)
		   Line[LinePos++] = Seg == DiagCol ? FontRow[3] : ' ';
		break;
         }
      for (Seg = 0; Seg < Spacing; Seg++)
         Line[LinePos++] = ' ';
      DiagCol += Upper ? 1 : -1;
   }
}

void WriteHoriz(char Line[], int LinePos, int CharLim[], char FontRow[],
int Spacing, int Size)
/*
Write segment row with only horizontal components (top, middle, and bottom rows)
char Line[],	  Line from buffer to write to
FontRow[];	  Font for row to write
int CharLim[],	  Which columns of font to write
LinePos,	  Position in Line to start writing to
  Number of display chars to write for segments with variable horizontal size
Size;
*/
{
   int FontCol,	/* Which column of font is being written */
   /* Counter for printing multiple display chars for segments with
      variable horizontal size */
   SegCount;
   char SegChar;	/* Char to write multiple times */

   /* Write cols required */
   for (FontCol = CharLim[0]; FontCol <= CharLim[1]; FontCol++)
      switch (FontCol) {
         case 0: case 2: case 4:
	     /* Store a non-repeating char */
	      Line[LinePos++] = FontRow[FontCol];
	      break;
         case 1: case 3:	/* Store a repeating char */
	      SegChar = FontRow[FontCol];  /* Find char to repeat */
	      /* Print char as many times as neccessary */
	      for (SegCount = 0; SegCount < Size; SegCount++)
	         Line[LinePos++] = SegChar;
	      break;
	 }
      for (SegCount = 0; SegCount < Spacing; SegCount++)
         Line[LinePos++] = ' ';
}

/*
Reads lines of input from InFile and prints them in segment font
FILE *InFile;	  Input file
int Backwards,
Size,		  Number of chars per segment to use
Width,		  Char width of output device
Spacing,	  Number of spaces between output chars
CharWidths[],	  Widths of chars when expanded
Dim[][2];	  First and last cols used by each char
CharDesc Font[];	  Font to use in printing
int UseTabs;	  Whether to use tabs instead of spaces whenever possible
char *S;	  String to print instead of a file
*/
void PrintAcross(FILE *InFile, int Backwards, int Size, int Width, 
CharDesc Font[], int Dim[][2], int Spacing, int CharWidths[], int UseTabs,
char *S)
{
   char InLine[MAX_LINE];
   int NumChars = 0;
   int NumRows,I;
   InBuf Buffer;
   char *OutBuf[MAX_COLS];

   /* Two * Size buffer rows for the variable-size seg char rows, 3 for the
      fixed size rows, and 1 buffer row for vertical spacing */
   NumRows = Size * 2 + 4;
   /* Set up memory for output buffer lines */
   for (I = 0; I < NumRows; I++)
      OutBuf[I] = (char *) CallocT(Width + Spacing,sizeof(char));
   Buffer.InFile = InFile;
   Buffer.CurChar = Buffer.LineBuf;
   if (!InFile) {
       strcpy(Buffer.LineBuf,S);
    }
    else
       Buffer.LineBuf[0] = '\0';
   Buffer.Backwards = Backwards;
   while ((!InFile && NumChars != -1) || (InFile && !feof(InFile))) {
      NumChars = GetChars(&Buffer,Width - 1,CharWidths,InLine,Spacing);
      if (NumChars > 0)
         PrintInRow(InLine,NumChars,Size,Font,Spacing,Dim,OutBuf,CharWidths,
          UseTabs);
   }
}

char NextChar(InBuf *Buffer)
{
   char C;
   char *NewLine;

   C = *(Buffer -> CurChar++);
   if (C == '\0') {
      if (!(Buffer->InFile) || feof(Buffer -> InFile) ||
       fgets(Buffer -> LineBuf,MAX_LINE,Buffer -> InFile) == NULL)
         return(EOF);
      if (Buffer -> Backwards) {
         NewLine = strchr(Buffer -> LineBuf,'\n');
         if (NewLine != NULL)
            *NewLine = '\0';
         strrev(Buffer -> LineBuf);
         if (NewLine != NULL)
            *NewLine = '\n';
      }
      Buffer -> CurChar = Buffer -> LineBuf;
      return(*(Buffer -> CurChar++));
   }
   return(C);
}

void RestoreChar(InBuf *Buffer)
{
   if (Buffer -> CurChar != Buffer -> LineBuf)
      Buffer -> CurChar--;
}

/* Gets chars using Buffer and puts them in Line.  Chars are gotten until
   eof is reached, a newline is found, or a character is gotten that would
   cause the sum of the widths of the characters gotten and the spaces between
   them to exceed Width.  Newlines are not included in Line.
Input buffer.  Allows entire lines of input to be read so that they can
   be reversed (if the -b flag is given) before being divided up into output
   lines.
Width of output device actually used (width given minus 1 to avoid line wrap)
int Width,
CharWidths[],	Width of output chars
Spacing;	Spaces between output chars
char Line[];	Line to put chars read in in
*/
int GetChars(InBuf *Buffer, int Width, int CharWidths[], char Line[],
int Spacing)
{
   int CharsRead = 0;	/* Number of chars read */
   char InChar;		/* Char read */

   while (Width > 0) {
      InChar = NextChar(Buffer);
      if (InChar == EOF || InChar == '\n')
         break;
      Width -= CharWidths[InChar];
      if (Width >= 0)
         Line[CharsRead++] = InChar;
      else	/* If no room for char put it back in buffer */
         RestoreChar(Buffer);
      Width -= Spacing;
   }
   if (Width <= 0 && NextChar(Buffer) != '\n')
      RestoreChar(Buffer);
   /* If line is a blank line, make it consist of a space so that it will not
      be taken to be a null line (line of 0 length because EOF was first char
      read) */
   if (InChar == '\n' && !CharsRead)
      Line[CharsRead++] = ' ';
   Line[CharsRead] = '\0';	/* Terminate line */
   if (InChar == EOF && !CharsRead)
      return(-1);
   else
      return(CharsRead);
}

/* If ConvertSpaces is true, replaces spaces in Line with a tab wherever
   possible (except replacement is not done if a single space would be replaced
   by a tab); then deletes trailing spaces and tabs.  All control chars in Line
   except tabs are taken to occupy one space. */
void CompressLine(char *Line, int ConvertSpaces)
{
   if (ConvertSpaces)
      SpcsToTabs(Line);
   DeleteEndWhite(Line);
}

/* Replaces spaces in Line with a tab wherever possible (except replacement is
   not done if a single space would be replaced by a tab).
   All control chars in Line except tabs are taken to occupy one space. */
void SpcsToTabs(char *Line)
{
   int CharCount = 0;	/* How many chars have been found since the last
   			   tab position */
   int I;		/* Source char being converted */
   int NewPos = -1;	/* Position in line that chars are being written to */

   for (I = 0; Line[I] != '\0'; I++) {
      CharCount++;
      if (Line[I] == '\t')	/* If char is a tab, delete any immediately
            preceding spaces within the tab block that it will skip */
         while (--CharCount && Line[NewPos] == ' ')
            NewPos--;
      Line[++NewPos] = Line[I];	/* Copy char to new position */
      if (CharCount == 8) {	/* If end of a tab block has been reached */
         if (Line[NewPos] == ' ' && Line[NewPos - 1] == ' ') {
            /* If at least 2 spaces would be replaced, delete trailing
               spaces in the tab block and write a tab instead */
            NewPos--;
            while (--CharCount && Line[--NewPos] == ' ')
               ;
            Line[++NewPos] = '\t';
         }
         CharCount = 0;	/* Reset char count */
      }
   }
   Line[++NewPos] = '\0';	/* Add terminator */
}

/* Deletes trailing spaces and tabs. */
void DeleteEndWhite(char *Line)
{
   char *LineEnd;

   for (LineEnd = strchr(Line,'\0') - 1;
    (*LineEnd == ' ' || *LineEnd == '\t') && LineEnd >= Line; LineEnd--)
      *LineEnd = '\0';
}

void SetInts(int Ints[],int Num, int Val)
{
   int I;

   for (I = 0; I < Num; I++)
      Ints[I] = Val;
}
